#include "StdAfx.h"

//////////////////////////////////////////////////////////////////////////

#include "NScrollBar.h"
#include <math.h>
#include "ScrollBarUI_Solid.h"

IMPLEMENT_RTTI(CNScrollBar, CContainer)
IMPLEMENT_RTTI(IScrollBarUI, IComponentUI)
IMPLEMENT_ROOT_RTTI(IScrollBarCustomize)
IMPLEMENT_RTTI(CDefScrollBarCustomize, IScrollBarCustomize)

CNScrollBar::CNScrollBar(IScrollBarUI* pNSBUI /*= NULL*/, 
						 DWORD dwStyle /*= NSB_STYLE_HSCROLL*/, 
						 int nButtonWidth /*= 0*/, 
						 int nThumbWidth /*= 0*/,
						 int nHandleWidth /*= 0*/, 
						 IScrollBarCustomize* pScrollBarCustomize /*= NULL*/	)
{
	m_nHandleHop = 1;
	m_dwMinTime = 10;
	m_nTimeMultiplyNum = 10;
	m_nHopMultiplyNum = 10;
	m_fHandleMoment = 0.0f;
	m_nMargin = 0;
	m_dwTimerStartTime = 0;
	m_dwTimerElasedTime = 0;

	if (pNSBUI == NULL)
	{
		m_pNSBUI = new CScrollBarUI_Solid();
	}
	else
	{
		ATLASSERT(IS_KIND_OF(IScrollBarUI, pNSBUI) == TRUE);
		if ( IS_KIND_OF(IScrollBarUI, pNSBUI) == FALSE ) return;
		
		m_pNSBUI = pNSBUI;
	}

	m_nBtnWidth = (nButtonWidth == 0) ? ::GetSystemMetrics(SM_CXHSCROLL) : nButtonWidth;
	m_nThumbWidth = (nThumbWidth == 0) ? ::GetSystemMetrics(SM_CXHSCROLL) : nThumbWidth;
	m_nHandleWidth = (nHandleWidth == 0) ? ::GetSystemMetrics(SM_CXHSCROLL) : nHandleWidth;

	m_bHScroll = (dwStyle & NSB_STYLE_HSCROLL) != NULL;
	m_bHandle = (dwStyle & NSB_STYLE_HANDLE) != NULL;
	m_ptMousePressed.SetPoint(0, 0);
	m_rcShaftSelected.SetRectEmpty();
	m_state = E_NSB_STATE_NONE;
	m_bEnable = TRUE;
	m_bForcedTimerStop = FALSE;
	m_bMouseCaptured = FALSE;
	
	m_si.cbSize = sizeof(SCROLLINFO);
	m_si.fMask = SIF_ALL;
	m_si.nMax = 100;
	m_si.nMin = 0;
	m_si.nPage = 0;
	m_si.nPos = 0;
	m_si.nTrackPos = 0;

	m_pLUPanel = new CN4StatePanel();
	m_pRDPanel = new CN4StatePanel();
	m_pThumbPanel = new CN4StatePanel();
	m_pHandlePanel = new CN4StatePanel();

	m_pHandlePanel->SetVisible(m_bHandle);

	Add(m_pLUPanel);
	Add(m_pRDPanel);
	Add(m_pThumbPanel);
	Add(m_pHandlePanel);

	m_pShaftListener = new CNScrollBarListener(this);
	m_pLeftUpPanelListener = new CNScrollBarLeftUpPanelListener(this);
	m_pRightDownPanelListener = new CNScrollBarRightDownPanelListener(this);
	m_pHandlePanelListener = new CNScrollBarHandlePanelListener(this);
	m_pThumbPanelListener = new CNScrollBarThumbPanelListener(this);

	m_pScrollBarCustomize = (pScrollBarCustomize == NULL) ?
							new CDefScrollBarCustomize(this) : pScrollBarCustomize;
	// [:2008-12-18] Memory Leak  : Ҵߴ üũ
	m_bAllocScrollBarCustom = (pScrollBarCustomize == NULL);

	AddMouseListener(m_pShaftListener);
	m_pLUPanel->AddMouseListener(m_pLeftUpPanelListener);
	m_pRDPanel->AddMouseListener(m_pRightDownPanelListener);
	m_pThumbPanel->AddMouseListener(m_pThumbPanelListener);
	
	if(m_bHandle)
	{
		m_pHandlePanel->AddMouseListener(m_pHandlePanelListener);
	}
	
	m_pLUPanel->SetUI(m_pNSBUI->GetLUPanelUI());
	m_pRDPanel->SetUI(m_pNSBUI->GetRDPanelUI());
	m_pThumbPanel->SetUI(m_pNSBUI->GetThumbPanelUI());
	m_pHandlePanel->SetUI(m_pNSBUI->GetHandlePanelUI());

	SetLayoutManager(this);
}

CNScrollBar::~CNScrollBar()
{
	if(m_pShaftListener)
	{
		delete m_pShaftListener;
		m_pShaftListener = NULL;
	}

	if(m_pLeftUpPanelListener)
	{
		delete m_pLeftUpPanelListener;
		m_pLeftUpPanelListener = NULL;
	}

	if(m_pRightDownPanelListener)
	{
		delete m_pRightDownPanelListener;
		m_pRightDownPanelListener = NULL;	
	}

	if(m_pThumbPanelListener)
	{
		delete m_pThumbPanelListener;
		m_pThumbPanelListener = NULL;
	}

	if(m_pHandlePanelListener)
	{
		delete m_pHandlePanelListener;
		m_pHandlePanelListener = NULL;
	}

	// [:2008-12-18] Memory Leak  : ҴǾٸ 
	if (m_pScrollBarCustomize && m_bAllocScrollBarCustom)
	{
		delete m_pScrollBarCustomize;
		m_pScrollBarCustomize = NULL;
	}
}

BOOL CNScrollBar::GetScrollInfo(LPSCROLLINFO lpScrollInfo, UINT nMask)
{
	lpScrollInfo->fMask = nMask;
	*lpScrollInfo = m_si;
	return TRUE;
}

int	CNScrollBar::GetScrollLimit()
{
	return m_si.nMax;
}

int	CNScrollBar::GetScrollPos( )
{
	return m_si.nPos;
}

void CNScrollBar::GetScrollRange(LPINT lpMinPos, LPINT lpMaxPos)
{
	*lpMinPos = m_si.nMin;
	*lpMaxPos = m_si.nMax;
}

int	CNScrollBar::SetScrollInfo(LPSCROLLINFO lpScrollInfo, BOOL bRedraw)
{	
	if((m_si.nMin != lpScrollInfo->nMin)
		|| (m_si.nMax != lpScrollInfo->nMax)
		|| (m_si.nPos != lpScrollInfo->nPos)
		|| (m_si.nPage != lpScrollInfo->nPage))
	{
		m_si = *lpScrollInfo;
		_RefreshPos();

		if(bRedraw)
		{
			InvalidateRect(NULL, FALSE);
		}
	}

	m_dwTimerElasedTime = ::GetTickCount() - m_dwTimerStartTime;
	return lpScrollInfo->nPos;
}

int	CNScrollBar::SetScrollPos(int nPos, BOOL bRedraw)
{
	m_si.nPos = nPos;

	_RefreshPos();
	
	if(bRedraw)
	{
		InvalidateRect(NULL, FALSE);
	}

	return m_si.nPos;
}

void CNScrollBar::SetScrollRange(int nMinPos, int nMaxPos, BOOL bRedraw)
{
	m_si.nMin = nMinPos;
	m_si.nMax = nMaxPos;

	_RefreshPos();
	
	if(bRedraw)
	{
		InvalidateRect(NULL, FALSE);
	}
}

void CNScrollBar::ShowScrollBar(BOOL bShow)
{
	SetVisible(bShow);
}

void CNScrollBar::AddScrollListener(IScrollListener* pScrollListener)
{
	ATLASSERT(pScrollListener != NULL);
	ATLASSERT(m_vActionListener.Find(pScrollListener) == NULL);
	m_vActionListener.AddTail(pScrollListener);
}

void CNScrollBar::RemoveScrollListener(IScrollListener* pScrollListener)
{
	ATLASSERT(pScrollListener != NULL);
	POSITION pos = m_vActionListener.GetHeadPosition();
	pos = m_vActionListener.Find(pScrollListener, pos);
	if (pos)
	{
		m_vActionListener.RemoveAt(pos);
	}
}

void CNScrollBar::FireScroll(CScrollEvent* pScrollEvent)
{
	if(this->IsWindowEnabled() == FALSE) return;

	POSITION pos = m_vActionListener.GetHeadPosition();
	while (pos)
	{
		IScrollListener* pActionListener = m_vActionListener.GetNext(pos);
		if (pActionListener)
		{
			if(m_bHScroll)
			{
				pActionListener->OnHScroll(pScrollEvent);
			}
			else
			{
				pActionListener->OnVScroll(pScrollEvent);
			}
		}
	}
}

BOOL CNScrollBar::OnTimer( DWORD dwTimerID)
{
	if(m_bForcedTimerStop)
	{
		m_Timer.StopTimer();
		return FALSE;
	}

	m_dwTimerStartTime = ::GetTickCount();
	m_Timer.StopTimer();

	switch(dwTimerID)
	{
	case NSB_TIMERID_HANDLE:
		{
			CRect rcHandle;
			m_pHandlePanel->GetBounds(&rcHandle);
			float fHopAccelation = 1 + (float)(m_nHopMultiplyNum - 1) * fabs(m_fHandleMoment);
			
			if(m_bHScroll)
			{
				if(rcHandle.left < m_rcMidHandle.left)
				{
					m_si.nPos = max((int)(m_si.nPos - m_nHandleHop * fHopAccelation), m_si.nMin);
				}
				else
				{
					m_si.nPos = min((int)(m_si.nPos + m_nHandleHop * fHopAccelation), m_si.nMax - (int)m_si.nPage);
				}
			}
			else
			{
				if(rcHandle.top < m_rcMidHandle.top)
				{
					m_si.nPos = max((int)(m_si.nPos - m_nHandleHop * fHopAccelation), m_si.nMin);
				}
				else
				{
					m_si.nPos = min((int)(m_si.nPos + m_nHandleHop * fHopAccelation), m_si.nMax - (int)m_si.nPage);
				}
			}
	
			m_si.nTrackPos = m_si.nPos;
			CScrollEvent scrollevent(SB_THUMBTRACK, m_si.nPos, this);
			FireScroll(&scrollevent);

			_RefreshPos();
			InvalidateRect(NULL);
		}
		break;
	case NSB_TIMERID_SHAFT:
		{
			CRect rcThumb;
			m_pThumbPanel->GetBounds(&rcThumb);

			if(m_bHandle == FALSE)
			{
				if(m_bHScroll)
				{
					if(m_ptMousePressed.x < rcThumb.left)
					{
						CScrollEvent scrollevent(SB_PAGELEFT, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.right = rcThumb.left;
					}
					else if(m_ptMousePressed.x > rcThumb.right)
					{
						CScrollEvent scrollevent(SB_PAGERIGHT, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.left = rcThumb.right;
					}
					else
					{
						m_Timer.StopTimer();
						m_pThumbPanel->ReleaseCapture();
						m_bMouseCaptured = FALSE;
						m_rcShaftSelected.SetRectEmpty();
						this->Invalidate();
						return TRUE;
					}
				}
				else
				{
					if(m_ptMousePressed.y < rcThumb.top)
					{
						CScrollEvent scrollevent(SB_PAGEUP, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.bottom = rcThumb.top;
					}
					else if(m_ptMousePressed.y > rcThumb.bottom)
					{
						CScrollEvent scrollevent(SB_PAGEDOWN, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.top = rcThumb.bottom;
					}
					else
					{
						// PM-719 2007/10/19 22:17:47 
						// PM-720 2007/10/19 22:19:35 
						/* {kr}
							ũѹ Body Ŭϸ Ÿ̸Ӱ ۵ǰ,
							Thumb 콺 ġ   Ÿ̸Ӱ Ǿ ϴµ.
							Ÿ̸Ӱ   ʰ    Thumb ž .
							 ó  .
						{/kr} */
						m_Timer.StopTimer();
						m_pThumbPanel->ReleaseCapture();
						m_bMouseCaptured = FALSE;
						m_rcShaftSelected.SetRectEmpty();
						this->Invalidate();
						return TRUE;
					}
				}
			}
			else
			{
				CRect rcNSB;
				GetClientRect(&rcNSB);
				CPoint ptMiddleNSB = rcNSB.CenterPoint();
				
				if(m_bHScroll == TRUE)
				{
					if(m_ptMousePressed.x < ptMiddleNSB.x)
					{
						CScrollEvent scrollevent(SB_PAGELEFT, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.right = rcThumb.left;
					}
					else if(m_ptMousePressed.x > ptMiddleNSB.x)
					{
						CScrollEvent scrollevent(SB_PAGERIGHT, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.left = rcThumb.right;
					}
					else
					{
						m_Timer.StopTimer();
						m_pThumbPanel->ReleaseCapture();
						m_rcShaftSelected.SetRectEmpty();
					}
				}
				else
				{
					if(m_ptMousePressed.y < ptMiddleNSB.y)
					{
						CScrollEvent scrollevent(SB_PAGEUP, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.bottom = rcThumb.top;
					}
					else if(m_ptMousePressed.y > ptMiddleNSB.y)
					{
						CScrollEvent scrollevent(SB_PAGEDOWN, m_si.nPos, this);
						FireScroll(&scrollevent);
						GetClientRect(&m_rcShaftSelected);
						m_rcShaftSelected.top = rcThumb.bottom;
					}
					else
					{
						m_Timer.StopTimer();
						m_pThumbPanel->ReleaseCapture();
						m_rcShaftSelected.SetRectEmpty();
					}
				}
			}
			Invalidate();
		}
		break;
	case NSB_TIMERID_LUBTN:
		{
			if(m_bHScroll)
			{
				CScrollEvent scrollevent(SB_LINELEFT, m_si.nPos, this);
				FireScroll(&scrollevent);
			}
			else
			{
				CScrollEvent scrollevent(SB_LINEUP, m_si.nPos, this);
				FireScroll(&scrollevent);
			}
		}
		break;
	case NSB_TIMERID_RDBTN:
		{
			if(m_bHScroll)
			{
				CScrollEvent scrollevent(SB_LINERIGHT, m_si.nPos, this);
				FireScroll(&scrollevent);
			}
			else
			{
				CScrollEvent scrollevent(SB_LINEDOWN, m_si.nPos, this);
				FireScroll(&scrollevent);
			}
		}
		break;
	}

	DWORD dwRemainTime = (DWORD)max(0, (int)m_dwCurrentTimerPeriod - (int)m_dwTimerElasedTime);

	m_Timer.SetTimer(dwRemainTime, 1, dwTimerID, this);
	
	return TRUE;
}

void CNScrollBar::SetHandleOption(DWORD dwMinTime /*= 10 milli seconds*/, int nHandleHop /* = 1 unit*/, 
								  int nTimeMultiplyNum /*= 10 times*/, int nHopMultiplyNum /*= 10 times*/)
{
	m_dwMinTime = dwMinTime;
	m_nHandleHop = nHandleHop;
	m_nTimeMultiplyNum = nTimeMultiplyNum;
	m_nHopMultiplyNum = nHopMultiplyNum;
}

void CNScrollBar::GetHandleOption(DWORD* pdwMinTime, int* pnHandleHop)
{
	*pdwMinTime = m_dwMinTime;
	*pnHandleHop = m_nHandleHop;
}

void CNScrollBar::SetUI(IComponentUI* pComponentUI)
{
	ATLASSERT(pComponentUI);
	if(!pComponentUI) return;

	m_pNSBUI = reinterpret_cast<IScrollBarUI*>(pComponentUI);

	m_pLUPanel->SetUI(m_pNSBUI->GetLUPanelUI());
	m_pRDPanel->SetUI(m_pNSBUI->GetRDPanelUI());
	m_pThumbPanel->SetUI(m_pNSBUI->GetThumbPanelUI());
	m_pHandlePanel->SetUI(m_pNSBUI->GetHandlePanelUI());
}

void CNScrollBar::ModifyStyle(DWORD dwStyle)
{
	m_bHScroll = (dwStyle & NSB_STYLE_HSCROLL) != NULL;
	m_bHandle = (dwStyle & NSB_STYLE_HANDLE) != NULL;

	m_pHandlePanel->SetVisible(m_bHandle);

	_RefreshPos();
	m_pHandlePanel->MoveWindow(&m_rcMidHandle);

	m_pHandlePanel->RemoveMouseListener(m_pHandlePanelListener);

	if(m_bHandle)
	{
		m_pHandlePanel->AddMouseListener(m_pHandlePanelListener);	
	}

	InvalidateRect(NULL);
}

void CNScrollBar::Enable(BOOL bEnable /*= TRUE*/)
{
	if(m_bEnable == bEnable) return;

	RemoveMouseListener(m_pShaftListener);
	m_pLUPanel->RemoveMouseListener(m_pLeftUpPanelListener);
	m_pRDPanel->RemoveMouseListener(m_pRightDownPanelListener);
	m_pThumbPanel->RemoveMouseListener(m_pThumbPanelListener);
	m_pHandlePanel->RemoveMouseListener(m_pHandlePanelListener);

	if(bEnable)
	{
		m_pLUPanel->AddMouseListener(m_pLeftUpPanelListener);
		m_pRDPanel->AddMouseListener(m_pRightDownPanelListener);
		m_pThumbPanel->AddMouseListener(m_pThumbPanelListener);
		AddMouseListener(m_pShaftListener);

		if(m_bHandle)
		{	
			m_pHandlePanel->AddMouseListener(m_pHandlePanelListener);
		}

		m_state = E_NSB_STATE_NONE;
		m_pLUPanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
		m_pRDPanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
		m_pThumbPanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
		m_pHandlePanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);

		InvalidateRect(NULL);
	}
	else
	{
		m_state = E_NSB_STATE_DISABLED;
		m_pLUPanel->SetState(CN4StatePanel::E_PANEL_4STATE_DIS);
		m_pRDPanel->SetState(CN4StatePanel::E_PANEL_4STATE_DIS);
		m_pThumbPanel->SetState(CN4StatePanel::E_PANEL_4STATE_DIS);
		m_pHandlePanel->SetState(CN4StatePanel::E_PANEL_4STATE_DIS);

		InvalidateRect(NULL);
	}

	m_bEnable = bEnable;
}

void CNScrollBar::LayoutContainer(CContainer* pContainer)
{
	_RefreshPos();
	if(m_bHandle)
	{
		m_pHandlePanel->MoveWindow(&m_rcMidHandle);
	}
}

void CNScrollBar::OnDraw(HDC hDC)
{
	if(m_bMouseCaptured == TRUE)
	{
		CRect rcNSB;
		GetClientRect(&rcNSB);
		CRect rcThumb;
		m_pThumbPanel->GetBounds(&rcThumb);
		m_rcShaftSelected.CopyRect(&rcNSB);

		if(m_bHScroll == TRUE)
		{
			if(m_bIsLeftUp == TRUE)
			{
				m_rcShaftSelected.right = rcThumb.left;
			}
			else
			{
				m_rcShaftSelected.left = rcThumb.right;
			}
		}
		else
		{
			if(m_bIsLeftUp == TRUE)
			{
				m_rcShaftSelected.bottom = rcThumb.top;
			}
			else
			{
				m_rcShaftSelected.top = rcThumb.bottom;
			}
		}
	}
	else
	{
		m_rcShaftSelected.SetRectEmpty();
	}

	m_pNSBUI->DrawBackground(hDC, this, &m_rcShaftSelected);
	__super::OnDraw(hDC);
}

void CNScrollBar::_RefreshPos()
{
	if((int)m_si.nPage < m_si.nMax - m_si.nMin)
	{
		if(m_bEnable == FALSE)
		{
			m_pThumbPanel->SetVisible(TRUE);

			if(m_bHandle)
			{
				m_pHandlePanel->SetVisible(TRUE);
			}

			Enable(TRUE);
		}
	}
	else
	{
		m_pThumbPanel->SetVisible(FALSE);
		m_pHandlePanel->SetVisible(FALSE);
		Enable(FALSE);	
	}

	CRect rcLU, rcRD, rcNSB, rcThumb;
	GetClientRect(&rcNSB);

	CRect rcBound;
	GetBounds(rcBound);

	if(m_bHScroll)
	{
		rcLU.left = m_nMargin;
		rcLU.top = 0;
		rcLU.right = rcLU.left + m_nBtnWidth;
		rcLU.bottom = rcLU.top + rcBound.Height();
		m_pLUPanel->MoveWindow(&rcLU);

		rcRD.right = rcNSB.right - m_nMargin;
		rcRD.left = rcRD.right - m_nBtnWidth;
		rcRD.top = 0;
		rcRD.bottom = rcRD.top + rcBound.Height();
		m_pRDPanel->MoveWindow(&rcRD);

		BOOL bIsEnable = IsEnable();
		if(bIsEnable == FALSE) return;

		rcThumb.CopyRect(&rcLU);

		if(m_si.nPage)
		{
			int nVariableThumbWidth = (rcRD.left - rcLU.right) * (int)m_si.nPage / max(1, m_si.nMax - m_si.nMin);

			if(nVariableThumbWidth > m_nThumbWidth / 2)
			{
				rcThumb.OffsetRect(rcLU.Width() + (rcRD.left - rcLU.right) * (m_si.nPos - m_si.nMin) 
					/ max(1, m_si.nMax - m_si.nMin), 0);
				rcThumb.right = rcThumb.left + nVariableThumbWidth;
			}
			else
			{
				nVariableThumbWidth = m_nThumbWidth / 2;
				rcThumb.OffsetRect(rcLU.Width() + (rcRD.left - rcLU.right - nVariableThumbWidth) * (m_si.nPos - m_si.nMin) 
					/ max(1, m_si.nMax - m_si.nMin - m_si.nPage), 0);
				rcThumb.right = rcThumb.left + nVariableThumbWidth;
			}
			
		}
		else
		{
			rcThumb.OffsetRect(rcLU.Width() + (rcRD.left - rcLU.right - m_nThumbWidth) * (m_si.nPos - m_si.nMin) 
							/ max(1, m_si.nMax - m_si.nMin), 0);
			rcThumb.right = rcThumb.left + m_nThumbWidth;
		}

		m_pThumbPanel->MoveWindow(&rcThumb);

		if(m_bHandle)
		{
			m_rcMidHandle.left = rcLU.right + (rcRD.left - rcLU.right) / 2 - m_nHandleWidth / 2;
			m_rcMidHandle.top = 0;
			m_rcMidHandle.right = m_rcMidHandle.left + m_nHandleWidth;
			m_rcMidHandle.bottom = m_rcMidHandle.top + rcNSB.Height();
		}
	}
	else
	{
		rcLU.top = m_nMargin;
		rcLU.left = 0;
		rcLU.bottom = rcLU.top + m_nBtnWidth;
		rcLU.right = rcLU.left + rcBound.Width();
		m_pLUPanel->MoveWindow(&rcLU);

		rcRD.bottom = rcNSB.bottom - m_nMargin;
		rcRD.top = rcRD.bottom - m_nBtnWidth;
		rcRD.left = 0;
		rcRD.right = rcRD.left + rcBound.Width();
		m_pRDPanel->MoveWindow(&rcRD);

		BOOL bIsEnable = IsEnable();
		if(bIsEnable == FALSE) return;

		rcThumb.CopyRect(&rcLU);

		if(m_si.nPage)
		{
			int nVariableThumbWidth = (rcRD.top - rcLU.bottom) * (int)m_si.nPage / max(1, m_si.nMax - m_si.nMin);

			if(nVariableThumbWidth > m_nThumbWidth / 2)
			{
				rcThumb.OffsetRect(0, rcLU.Height() + (rcRD.top - rcLU.bottom) * 
					(m_si.nPos - m_si.nMin) / max(1, m_si.nMax - m_si.nMin));
				rcThumb.bottom = rcThumb.top + nVariableThumbWidth;
			}
			else
			{
				nVariableThumbWidth = m_nThumbWidth / 2;
				rcThumb.OffsetRect(0, rcLU.Height() + (rcRD.top - rcLU.bottom - nVariableThumbWidth) * 
					(m_si.nPos - m_si.nMin) / max(1, m_si.nMax - m_si.nMin - m_si.nPage));
				rcThumb.bottom = rcThumb.top + nVariableThumbWidth;
			}
		}
		else
		{
			rcThumb.OffsetRect(0, rcLU.Height() + (rcRD.top - rcLU.bottom - m_nThumbWidth) * 
				(m_si.nPos - m_si.nMin) / max(1, m_si.nMax - m_si.nMin));
			rcThumb.bottom = rcThumb.top + m_nThumbWidth;
		}

		m_pThumbPanel->MoveWindow(&rcThumb);

		if(m_bHandle)
		{
			m_rcMidHandle.top = rcLU.bottom + (rcRD.top - rcLU.bottom) / 2 - m_nHandleWidth / 2;
			m_rcMidHandle.left = 0;
			m_rcMidHandle.bottom = m_rcMidHandle.top + m_nHandleWidth;
			m_rcMidHandle.right = m_rcMidHandle.left + rcNSB.Width();
		}
	}
}

void CNScrollBarListener::OnMousePressed(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_ptMousePressed = pMouseEvent->m_Point;

	CRect rcNSB;
	m_pNSB->GetClientRect(&rcNSB);
	CPoint ptMidNSB = rcNSB.CenterPoint();

	if(m_pNSB->m_bHScroll == TRUE)
	{
		m_pNSB->m_bIsLeftUp = (pMouseEvent->m_Point.x < ptMidNSB.x);
	}
	else
	{
		m_pNSB->m_bIsLeftUp = (pMouseEvent->m_Point.y < ptMidNSB.y);
	}

	m_pNSB->m_pScrollBarCustomize->OnShaftDown(m_pNSB->m_bIsLeftUp);

	m_pNSB->m_bMouseCaptured = TRUE;
	m_pNSB->m_pThumbPanel->SetCapture();
	
	m_pNSB->m_state = CNScrollBar::E_NSB_STATE_SELECTED;
	m_pNSB->InvalidateRect(NULL);
} 

void CNScrollBarListener::OnMouseReleased(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnShaftUp();

	m_pNSB->m_bMouseCaptured = FALSE;
	m_pNSB->m_pThumbPanel->ReleaseCapture();
	ATLTRACE(_T("CNScrollBarListener::OnMouseReleased ReleaseCapture \n")); //hhdtrace

	m_pNSB->m_state = CNScrollBar::E_NSB_STATE_NONE;
	m_pNSB->InvalidateRect(NULL);
}

void CNScrollBarListener::OnMouseEntered(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarListener::OnMouseExited(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarListener::OnMouseMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarListener::OnMouseWheelMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarLeftUpPanelListener::OnMousePressed(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnLeftUpPanelDown();
	
	m_pNSB->m_pLUPanel->SetCapture();

	m_pNSB->m_pLUPanel->SetState(CN4StatePanel::E_PANEL_4STATE_DOWN);
	m_pNSB->InvalidateRect(NULL);
} 

void CNScrollBarLeftUpPanelListener::OnMouseReleased(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnLeftUpPanelUp();

	m_pNSB->m_pLUPanel->ReleaseCapture();
	
	m_pNSB->m_pLUPanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
	m_pNSB->InvalidateRect(NULL);
}

void CNScrollBarLeftUpPanelListener::OnMouseEntered(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarLeftUpPanelListener::OnMouseExited(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_Timer.StopTimer();
}

void CNScrollBarLeftUpPanelListener::OnMouseMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarLeftUpPanelListener::OnMouseWheelMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarRightDownPanelListener::OnMousePressed(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnRightDownPanelDown();
	
	m_pNSB->m_pRDPanel->SetCapture();
	
	m_pNSB->m_pRDPanel->SetState(CN4StatePanel::E_PANEL_4STATE_DOWN);
	m_pNSB->InvalidateRect(NULL);
} 

void CNScrollBarRightDownPanelListener::OnMouseReleased(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnRightDownPanelUp();

	m_pNSB->m_pRDPanel->ReleaseCapture();

	m_pNSB->m_pRDPanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
	m_pNSB->InvalidateRect(NULL);
}

void CNScrollBarRightDownPanelListener::OnMouseEntered(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarRightDownPanelListener::OnMouseExited(CMouseEvent* pMouseEvent)
{	
	m_pNSB->m_Timer.StopTimer();
}

void CNScrollBarRightDownPanelListener::OnMouseMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarRightDownPanelListener::OnMouseWheelMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarThumbPanelListener::OnMousePressed(CMouseEvent* pMouseEvent)
{
	if(m_pNSB->m_bHandle == TRUE)
	{
		CRect rcThumb;
		m_pNSB->m_pThumbPanel->GetBounds(&rcThumb);
		pMouseEvent->m_Point.Offset(rcThumb.left, rcThumb.top);
		m_pNSB->m_pShaftListener->OnMousePressed(pMouseEvent);
	}
	else
	{
		m_bMousePressed = TRUE;
		m_ptOld = pMouseEvent->m_Point;
		m_pNSB->m_pThumbPanel->SetCapture();
		m_pNSB->m_pThumbPanel->SetState(CN4StatePanel::E_PANEL_4STATE_DOWN);
		m_pNSB->InvalidateRect(NULL);
	}
} 

void CNScrollBarThumbPanelListener::OnMouseReleased(CMouseEvent* pMouseEvent)
{
	if(m_pNSB->m_bHandle == TRUE)
	{
		CRect rcThumb;
		m_pNSB->m_pThumbPanel->GetBounds(&rcThumb);
		pMouseEvent->m_Point.Offset(rcThumb.left, rcThumb.top);
		m_pNSB->m_pShaftListener->OnMouseReleased(pMouseEvent);
	}
	else
	{
		m_bMousePressed = FALSE;
		m_pNSB->m_pThumbPanel->ReleaseCapture();
		ATLTRACE(_T("CNScrollBarThumbPanelListener::OnMouseReleased ReleaseCapture \n")); //hhdtrace
		m_pNSB->m_state = CNScrollBar::E_NSB_STATE_NONE;
		// PM-878 2007/11/02 14:36:46 
		/* {kr}
			 콺  ö ¿ mouse release ϸ, 
			mouse over  UI ׷  ֵ .
		{/kr} */
		if (m_pNSB->m_pThumbPanel->IsMouseOver() == TRUE)
		{
			m_pNSB->m_pThumbPanel->SetState(CN4StatePanel::E_PANEL_4STATE_OVER);
		}
		else
		{
			m_pNSB->m_pThumbPanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
		}
		m_pNSB->InvalidateRect(NULL);
	}
}

void CNScrollBarThumbPanelListener::OnMouseEntered(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarThumbPanelListener::OnMouseExited(CMouseEvent* pMouseEvent)
{	
}

void CNScrollBarThumbPanelListener::OnMouseMoved(CMouseEvent* pMouseEvent)
{
	if(!m_bMousePressed) return;

	int nXMove = pMouseEvent->m_Point.x - m_ptOld.x;
	int nYMove = pMouseEvent->m_Point.y - m_ptOld.y;

	CRect rcThumb, rcLU, rcRD;
	m_pNSB->m_pThumbPanel->GetBounds(&rcThumb);
	m_pNSB->m_pLUPanel->GetBounds(&rcLU);
	m_pNSB->m_pRDPanel->GetBounds(&rcRD);
	
	if(m_pNSB->m_bHScroll)
	{
		int nThumbWidth = rcThumb.Width();

		if(nXMove > 0)
		{
			rcThumb.right = min(rcRD.left, rcThumb.right + nXMove);
			rcThumb.left = rcThumb.right - nThumbWidth;
		}
		else
		{
			rcThumb.left = max(rcLU.right, rcThumb.left + nXMove);
			rcThumb.right = rcThumb.left + nThumbWidth;
		}
	}
	else
	{
		int nThumbHeight = rcThumb.Height();

		if(nYMove > 0)
		{
			rcThumb.bottom = min(rcRD.top, rcThumb.bottom + nYMove);
			rcThumb.top = rcThumb.bottom - nThumbHeight;
		}
		else
		{
			rcThumb.top = max(rcLU.bottom, rcThumb.top + nYMove);
			rcThumb.bottom = rcThumb.top + nThumbHeight;
		}
	}

	m_pNSB->m_pThumbPanel->MoveWindow(&rcThumb, TRUE);

	SCROLLINFO si;
	si = m_pNSB->m_si;

	if(m_pNSB->m_bHScroll)
	{
		if(si.nPage)
		{
			si.nPos = (si.nMax - si.nMin) * (rcThumb.left - rcLU.right) / (rcRD.left - rcLU.right); 
		}
		else
		{
			si.nPos = (si.nMax - si.nMin) * (rcThumb.left - rcLU.right) / (rcRD.left - rcLU.right - rcThumb.Width());
		}
	}
	else
	{
		if(si.nPage)
		{
			si.nPos = (si.nMax - si.nMin) * (rcThumb.top - rcLU.bottom) / (rcRD.top - rcLU.bottom); 
		}
		else
		{
			si.nPos = (si.nMax - si.nMin) * (rcThumb.top - rcLU.bottom) / (rcRD.top - rcLU.bottom - rcThumb.Height());
		}
	}

	m_pNSB->m_si.nTrackPos = m_pNSB->m_si.nPos;
	CScrollEvent scrollevent(SB_THUMBTRACK, si.nPos, m_pNSB);
	m_pNSB->FireScroll(&scrollevent);

	m_pNSB->m_si = si;
	m_pNSB->InvalidateRect(NULL, FALSE);
}

void CNScrollBarThumbPanelListener::OnMouseWheelMoved(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarHandlePanelListener::OnMousePressed(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnHandlePanelDown();
	
	m_bMousePressed = TRUE;
	m_ptOld = pMouseEvent->m_Point;
	
	m_pNSB->m_pHandlePanel->SetCapture();
	
	m_pNSB->m_pHandlePanel->SetState(CN4StatePanel::E_PANEL_4STATE_DOWN);
	m_pNSB->InvalidateRect(NULL);
} 

void CNScrollBarHandlePanelListener::OnMouseReleased(CMouseEvent* pMouseEvent)
{
	m_pNSB->m_pScrollBarCustomize->OnHandlePanelUp();
	
	m_bMousePressed = FALSE;
	m_pNSB->m_pHandlePanel->MoveWindow(&(m_pNSB->m_rcMidHandle));
	
	m_pNSB->m_pHandlePanel->ReleaseCapture();
	
	m_pNSB->m_pHandlePanel->SetState(CN4StatePanel::E_PANEL_4STATE_NORMAL);
	m_pNSB->InvalidateRect(NULL);
}

void CNScrollBarHandlePanelListener::OnMouseEntered(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarHandlePanelListener::OnMouseExited(CMouseEvent* pMouseEvent)
{
}

void CNScrollBarHandlePanelListener::OnMouseMoved(CMouseEvent* pMouseEvent)
{
	if(m_bMousePressed)
	{
		CRect rcHandle, rcLU, rcRD;
		m_pNSB->m_pHandlePanel->GetBounds(&rcHandle);
		m_pNSB->m_pLUPanel->GetBounds(&rcLU);
		m_pNSB->m_pRDPanel->GetBounds(&rcRD);

		int nXMove = pMouseEvent->m_Point.x - m_ptOld.x;
		int nYMove = pMouseEvent->m_Point.y - m_ptOld.y;

		if(m_pNSB->m_bHScroll)
		{	
			rcHandle.OffsetRect(nXMove, 0);
			if(rcHandle.left < rcLU.right)
			{
				rcHandle.left = rcLU.right;
				rcHandle.right = rcHandle.left + m_pNSB->m_nHandleWidth;
			}
			else if(rcHandle.right > rcRD.left)
			{
				rcHandle.right = rcRD.left;
				rcHandle.left = rcHandle.right - m_pNSB->m_nHandleWidth;
			}

			if(nXMove)
			{
				m_pNSB->m_fHandleMoment = (float)(rcHandle.left - m_pNSB->m_rcMidHandle.left) 
					/ (m_pNSB->m_rcMidHandle.left - rcLU.right);

				m_pNSB->m_pScrollBarCustomize->OnHandlePanelMove();
			}
		}
		else
		{
			rcHandle.OffsetRect(0, nYMove);
			if(rcHandle.top < rcLU.bottom)
			{
				rcHandle.top = rcLU.bottom;
				rcHandle.bottom = rcHandle.top + m_pNSB->m_nHandleWidth;
			}
			else if(rcHandle.bottom > rcRD.top)
			{
				rcHandle.bottom = rcRD.top;
				rcHandle.top = rcHandle.bottom - m_pNSB->m_nHandleWidth;
			}

			if(nYMove)
			{
				m_pNSB->m_fHandleMoment = (float)(rcHandle.top - m_pNSB->m_rcMidHandle.top)
					/ (m_pNSB->m_rcMidHandle.top - rcLU.bottom);

				m_pNSB->m_pScrollBarCustomize->OnHandlePanelMove();
			}
		}

		m_pNSB->m_pHandlePanel->MoveWindow(&rcHandle);
		m_pNSB->InvalidateRect(NULL, FALSE);
	}
}

void CNScrollBarHandlePanelListener::OnMouseWheelMoved(CMouseEvent* pMouseEvent)
{
}

void CDefScrollBarCustomize::OnShaftDown(BOOL bIsLeftUp)
{
	m_pNSB->m_bForcedTimerStop = FALSE;
	m_pNSB->m_dwCurrentTimerPeriod = m_pNSB->m_dwMinTime;	
	m_pNSB->m_Timer.SetTimer(0, 1, NSB_TIMERID_SHAFT, m_pNSB);
}

void CDefScrollBarCustomize::OnShaftUp()
{
	m_pNSB->m_bForcedTimerStop = TRUE;
	m_pNSB->m_Timer.StopTimer();
}

void CDefScrollBarCustomize::OnLeftUpPanelDown()
{
	m_pNSB->m_bForcedTimerStop = FALSE;
	m_pNSB->m_dwCurrentTimerPeriod = m_pNSB->m_dwMinTime;
	m_pNSB->m_Timer.SetTimer(0, 1, NSB_TIMERID_LUBTN, m_pNSB);
}

void CDefScrollBarCustomize::OnLeftUpPanelUp()
{
	m_pNSB->m_bForcedTimerStop = TRUE;
	m_pNSB->m_Timer.StopTimer();
}

void CDefScrollBarCustomize::OnRightDownPanelDown()
{
	m_pNSB->m_bForcedTimerStop = FALSE;
	m_pNSB->m_dwCurrentTimerPeriod = m_pNSB->m_dwMinTime;
	m_pNSB->m_Timer.SetTimer(0, 1, NSB_TIMERID_RDBTN, m_pNSB);
}

void CDefScrollBarCustomize::OnRightDownPanelUp()
{
	m_pNSB->m_bForcedTimerStop = TRUE;
	m_pNSB->m_Timer.StopTimer();
}

void CDefScrollBarCustomize::OnHandlePanelDown()
{
	m_pNSB->m_bForcedTimerStop = FALSE;
	m_bIsFirstMovement = TRUE;
}

void CDefScrollBarCustomize::OnHandlePanelUp()
{
	m_pNSB->m_bForcedTimerStop = TRUE;
	m_pNSB->m_Timer.StopTimer();
}

void CDefScrollBarCustomize::OnHandlePanelMove()
{
	m_pNSB->m_dwCurrentTimerPeriod = (DWORD)(float)(m_pNSB->m_nTimeMultiplyNum - 
						(float)(m_pNSB->m_nTimeMultiplyNum - 1) * fabs(m_pNSB->m_fHandleMoment)) * m_pNSB->m_dwMinTime;

	if(m_bIsFirstMovement == TRUE)
	{
		m_pNSB->m_dwTimerStartTime = ::GetTickCount();
		m_bIsFirstMovement = FALSE;
	}

	int nWaitTime = m_pNSB->m_dwCurrentTimerPeriod - ::GetTickCount() + m_pNSB->m_dwTimerStartTime;

	if(nWaitTime < 0)
	{
		nWaitTime = 0;
	}

	m_pNSB->m_Timer.StopTimer();

	m_pNSB->m_Timer.SetTimer(nWaitTime, 0, NSB_TIMERID_HANDLE, m_pNSB);
}